/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.guard.database.internal;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Properties;
import com.openexchange.database.Assignment;
import com.openexchange.database.AssignmentFactory;
import com.openexchange.database.DBPoolingExceptionCodes;
import com.openexchange.database.DatabaseService;
import com.openexchange.exception.OXException;
import com.openexchange.guard.configuration.GuardConfigurationService;
import com.openexchange.guard.configuration.GuardProperty;
import com.openexchange.guard.database.GuardDatabaseService;
import com.openexchange.guard.database.GuardShardingService;
import com.openexchange.guard.database.exception.GuardDatabaseExceptionCodes;
import com.openexchange.guard.database.utils.GuardConnectionWrapper;

/**
 * {@link GuardDatabaseServiceImpl}
 *
 * @author <a href="mailto:martin.schneider@open-xchange.com">Martin Schneider</a>
 * @since 2.4.0
 */
public class GuardDatabaseServiceImpl implements GuardDatabaseService {

    private static final org.slf4j.Logger LOG = org.slf4j.LoggerFactory.getLogger(GuardDatabaseServiceImpl.class);

    private final DatabaseService databaseService;
    private final GuardConfigurationService guardConfigurationService;
    private final AssignmentFactory assignmentFactory;

    public GuardDatabaseServiceImpl(DatabaseService databaseService, GuardConfigurationService guardConfigurationService, AssignmentFactory assignmentFactory) {
        this.databaseService = databaseService;
        this.guardConfigurationService = guardConfigurationService;
        this.assignmentFactory = assignmentFactory;
    }

    @Override
    public void reloadAssignments() throws OXException {
        this.assignmentFactory.reload();
    }

    //----------- read/write ox guard table -------------
    @Override
    public Connection getWritableForGuardNoTimeout() throws OXException {
        String guardSchemaName = this.guardConfigurationService.getProperty(GuardProperty.dbSchemaBaseName);
        Assignment assignment = assignmentFactory.get(guardSchemaName);
        if (assignment == null) {
            throw GuardDatabaseExceptionCodes.ASSIGNMENT_NOT_AVAILABLE_FOR_SCHEMA.create(guardSchemaName);
        }
        return databaseService.getWritable(assignment, true);
    }

    @Override
    public Connection getWritableForGuard() throws OXException {
        String guardSchemaName = this.guardConfigurationService.getProperty(GuardProperty.dbSchemaBaseName);
        Assignment assignment = assignmentFactory.get(guardSchemaName);
        if (assignment == null) {
            throw GuardDatabaseExceptionCodes.ASSIGNMENT_NOT_AVAILABLE_FOR_SCHEMA.create(guardSchemaName);
        }
        return databaseService.getWritable(assignment, false);
    }

    @Override
    public Connection getReadOnlyForGuard() throws OXException {
        String guardSchemaName = this.guardConfigurationService.getProperty(GuardProperty.dbSchemaBaseName);
        Assignment assignment = assignmentFactory.get(guardSchemaName);
        if (assignment == null) {
            throw GuardDatabaseExceptionCodes.ASSIGNMENT_NOT_AVAILABLE_FOR_SCHEMA.create(guardSchemaName);
        }
        return databaseService.getReadOnly(assignment, false);
    }

    @Override
    public void backWritableForGuardNoTimeout(Connection connection) {
        back(connection);
    }

    @Override
    public void backReadOnlyForGuard(Connection connection) {
        back(connection);
    }

    @Override
    public void backWritableForGuard(Connection connection) {
        back(connection);
    }

    //----------- read/write shard table -------------

    @Override
    public Connection getWritableForShardNoTimeout(int shardId) throws OXException {
        if (shardId == 0) {
            throw GuardDatabaseExceptionCodes.DB_ERROR.create("Shard with id 0 unknown!");
        }
        String shardName = getShardName(shardId);
        Assignment assignment = assignmentFactory.get(shardName);
        if (assignment == null) {
            throw GuardDatabaseExceptionCodes.ASSIGNMENT_NOT_AVAILABLE_FOR_SCHEMA.create(shardName);
        }
        return databaseService.getWritable(assignment, true);
    }

    @Override
    public Connection getWritableForShard(int shardId) throws OXException {
        if (shardId == 0) {
            throw GuardDatabaseExceptionCodes.DB_ERROR.create("Shard with id 0 unknown!");
        }
        String shardName = getShardName(shardId);
        Assignment assignment = assignmentFactory.get(shardName);
        if (assignment == null) {
            throw GuardDatabaseExceptionCodes.ASSIGNMENT_NOT_AVAILABLE_FOR_SCHEMA.create(shardName);
        }
        return databaseService.getWritable(assignment, false);
    }

    @Override
    public Connection getReadOnlyForShard(int shardId) throws OXException {
        if (shardId == 0) {
            throw GuardDatabaseExceptionCodes.DB_ERROR.create("Shard with id 0 unknown!");
        }
        String shardName = getShardName(shardId);
        Assignment assignment = assignmentFactory.get(shardName);
        if (assignment == null) {
            throw GuardDatabaseExceptionCodes.ASSIGNMENT_NOT_AVAILABLE_FOR_SCHEMA.create(shardName);
        }
        return databaseService.getReadOnly(assignment, false);
    }

    @Override
    public void backWritableForShardNoTimeout(int shardId, Connection connection) {
        back(connection);
    }

    @Override
    public void backReadOnlyForShard(int shardId, Connection connection) {
        back(connection);
    }

    @Override
    public void backWritableForShard(int shardId, Connection connection) {
        back(connection);
    }

    @Override
    public String getShardName(int shardId) {
        return guardConfigurationService.getProperty(GuardProperty.dbSchemaBaseName) + "_" + shardId;
    }

    @Override
    public String getConnectionUrl(String hostname) {
        return "jdbc:mysql://" + hostname;
    }

    private void back(Connection connection) {
        if (null == connection) {
            LOG.error("", DBPoolingExceptionCodes.NULL_CONNECTION.create());
            return;
        }
        try {
            connection.close();
        } catch (SQLException e) {
            LOG.error("", DBPoolingExceptionCodes.SQL_ERROR.create(e, e.getMessage()));
        }
    }

    //----------- generic connection retrieval -------------

    @Override
    public GuardConnectionWrapper getWritable(int userId, int contextId, int shard) throws OXException {
        if (contextId >= 0) {
            if (userId < 0) {
                return new GuardConnectionWrapperImpl(this.getWritableForGuard(), userId, contextId, shard);
            }
            return new GuardConnectionWrapperImpl(this.getWritable(contextId), userId, contextId, shard);
        } else {
            if (shard == 0) {
                shard = Services.getService(GuardShardingService.class).getShard(userId, contextId);
            }
            return new GuardConnectionWrapperImpl(this.getWritableForShard(shard), userId, contextId, shard);
        }
    }

    @Override
    public void backWritable(GuardConnectionWrapper connectionWrapper) {
        if (connectionWrapper.getContextId() >= 0) {
            if (connectionWrapper.getUserId() < 0) {
                this.backWritableForGuard(connectionWrapper.getConnection());
            } else {
                this.backWritable(connectionWrapper.getContextId(), connectionWrapper.getConnection());
            }
        } else {
            this.backWritableForShard(connectionWrapper.getShardId(), connectionWrapper.getConnection());
        }
    }

    @Override
    public GuardConnectionWrapper getReadOnly(int userId, int contextId, int shard) throws OXException {
        if (contextId >= 0) {
            if (userId < 0) {
                return new GuardConnectionWrapperImpl(this.getReadOnlyForGuard(), userId, contextId, shard);
            }
            return new GuardConnectionWrapperImpl(this.getReadOnly(contextId), userId, contextId, shard);
        } else {
            if (shard == 0) {
                shard = Services.getService(GuardShardingService.class).getShard(userId, contextId);
            }
            return new GuardConnectionWrapperImpl(this.getReadOnlyForShard(shard), userId, contextId, shard);
        }
    }

    @Override
    public void backReadOnly(GuardConnectionWrapper connectionWrapper) {
        if (connectionWrapper.getContextId() >= 0) {
            if (connectionWrapper.getUserId() < 0) {
                this.backReadOnlyForGuard(connectionWrapper.getConnection());
            } else {
                this.backReadOnly(connectionWrapper.getContextId(), connectionWrapper.getConnection());
            }
        } else {
            this.backReadOnlyForShard(connectionWrapper.getShardId(), connectionWrapper.getConnection());
        }
    }

    //----------- methods for INIT -------------

    @Override
    public Connection getWritableForShardInit() throws OXException {
        Properties properties = new Properties();
        int conntectionTimeout = this.guardConfigurationService.getIntProperty(GuardProperty.oxGuardShardDatabaseInitConnectionTimeout);
        properties.setProperty("connectTimeout", Integer.toString(conntectionTimeout));

        return getWritableForInit(this.guardConfigurationService.getProperty(GuardProperty.oxGuardShardDatabase), properties);
    }

    @Override
    public Connection getWritableForGuardInit() throws OXException {
        Properties properties = new Properties();
        int conntectionTimeout = this.guardConfigurationService.getIntProperty(GuardProperty.oxGuardDatabaseInitConnectionTimeout);
        properties.setProperty("connectTimeout", Integer.toString(conntectionTimeout));

        return getWritableForInit(this.guardConfigurationService.getProperty(GuardProperty.oxGuardDatabaseHostname), properties);
    }

    private Connection getWritableForInit(String host, Properties properties) throws OXException {
        String connectionString = "jdbc:mysql://" + host;

        properties.setProperty("user", this.guardConfigurationService.getProperty(GuardProperty.databaseUsername));
        properties.setProperty("password", this.guardConfigurationService.getProperty(GuardProperty.databasePassword));
        try {
            return DriverManager.getConnection(connectionString, properties);
        } catch (SQLException e) {
            throw GuardDatabaseExceptionCodes.CONNECTION_RETRIEVAL_ERROR.create(e, connectionString);
        }
    }

    @Override
    public void backWritableForInit(Connection connection) {
        back(connection);
    }

    //----------- delegates are ok -------------
    @Override
    public Connection getWritable(int contextId) throws OXException {
        return databaseService.getWritable(contextId);
    }

    @Override
    public Connection getReadOnly(int contextId) throws OXException {
        return databaseService.getReadOnly(contextId);
    }

    @Override
    public void backReadOnly(int contextId, Connection connection) {
        databaseService.backReadOnly(contextId, connection);
    }

    @Override
    public void backWritable(int contextId, Connection connection) {
        databaseService.backWritable(contextId, connection);
    }

    @Override
    public Connection getWritableForConfigDB() throws OXException {
        return databaseService.getWritable();
    }

    @Override
    public Connection getReadOnlyForConfigDB() throws OXException {
        return databaseService.getReadOnly();
    }

    @Override
    public void backReadOnlyForConfigDB(Connection connection) {
        databaseService.backReadOnly(connection);
    }

    @Override
    public void backWritableForConfigDB(Connection connection) {
        databaseService.backWritable(connection);
    }
}
